"""Plot source space NDVars with mayavi/pysurfer."""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>

from ._brain import (brain, cluster, dspm, surfer_brain, annot, bin_table,
                     dspm_bin_table, copy, p_map, annot_legend, butterfly)
